# (c) 2022-2025 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

import contextlib
import functools
import importlib
import importlib.util
import os
import sys
import sysconfig
import typing

from pathlib import Path

from gpep517.utils import get_toml, logger


@contextlib.contextmanager
def disable_zip_compression():
    import zipfile
    orig_open = zipfile.ZipFile.open
    orig_write = zipfile.ZipFile.write
    orig_writestr = zipfile.ZipFile.writestr

    @functools.wraps(zipfile.ZipFile.open)
    def override_open(self, name, mode="r", pwd=None,
                      *, force_zip64=False):
        if mode == "w":
            if not isinstance(name, zipfile.ZipInfo):
                name = zipfile.ZipInfo(name)
            name.compress_type = zipfile.ZIP_STORED
        ret = orig_open(self, name, mode, pwd, force_zip64=force_zip64)
        return ret

    @functools.wraps(zipfile.ZipFile.write)
    def override_write(self, filename, arcname=None,
                       compress_type=None, compresslevel=None):
        return orig_write(self, filename, arcname, zipfile.ZIP_STORED)

    @functools.wraps(zipfile.ZipFile.writestr)
    def override_writestr(self, zinfo_or_arcname, data,
                          compress_type=None, compresslevel=None):
        return orig_writestr(self, zinfo_or_arcname, data,
                             zipfile.ZIP_STORED)

    zipfile.ZipFile.open = override_open
    zipfile.ZipFile.write = override_write
    zipfile.ZipFile.writestr = override_writestr

    try:
        yield
    finally:
        zipfile.ZipFile.open = orig_open
        zipfile.ZipFile.write = orig_write
        zipfile.ZipFile.writestr = orig_writestr


@contextlib.contextmanager
def patch_sysconfig(sysroot: Path,
                    prefix: typing.Optional[Path],
                    ):
    get_vars = {}
    if prefix is not None:
        get_vars["installed_base"] = prefix

    stdlib_path = Path(sysconfig.get_path("stdlib", vars=get_vars))
    if not stdlib_path.is_absolute():
        raise RuntimeError(f"stdlib path {stdlib_path} is not absolute")
    sysroot_stdlib = sysroot / stdlib_path.relative_to("/")
    logger.info(f"Searching for sysconfig in {sysroot_stdlib}")
    data_paths = list(sysroot_stdlib.glob("_sysconfigdata_*.py"))
    if len(data_paths) != 1:
        raise RuntimeError(
            f"should have found one _sysconfigdata file, found {data_paths}")
    logger.info(f"Using sysconfig from {data_paths[0]}")

    data_spec = importlib.util.spec_from_file_location(
        "_sysconfig_data", data_paths[0])
    data_mod = importlib.util.module_from_spec(data_spec)
    data_spec.loader.exec_module(data_mod)
    sysroot_vars = data_mod.build_time_vars

    orig_config_vars = sysconfig.get_config_vars
    orig_get_platform = sysconfig.get_platform

    def patched_config_vars():
        cvars = orig_config_vars().copy()

        # path variables: we copy them from sysroot, and prepend sysroot
        for modvar in ("CONFINCLUDEDIR", "CONFINCLUDEPY", "INCLUDEDIR",
                       "INCLUDEPY", "LIBDIR"):
            srcvar = modvar
            if srcvar.startswith("CONF") and srcvar not in sysroot_vars:
                # PyPy does not define CONFINCLUDE*, distutils.sysconfig
                # works around that by hacking the path around manually.
                # However, setting CONFINCLUDE* here disables the hack
                # and gets sane behavior back.
                srcvar = srcvar[4:]
            if srcvar in sysroot_vars:
                sysroot_path = Path(sysroot_vars[srcvar])
                if not sysroot_path.is_absolute():
                    raise RuntimeError(
                        f"{srcvar} path {sysroot_path} is not absolute")
                cvars[modvar] = str(sysroot / sysroot_path.relative_to("/"))

        # ABI-specific variables: plain copy
        for modvar in ("SOABI", "EXT_SUFFIX"):
            if modvar in sysroot_vars:
                cvars[modvar] = sysroot_vars[modvar]

        return cvars

    def patched_get_platform():
        return sysroot_vars["MULTIARCH"]

    sysconfig.get_config_vars = patched_config_vars
    sysconfig.get_platform = patched_get_platform
    try:
        yield
    finally:
        sysconfig.get_config_vars = orig_config_vars
        sysconfig.get_platform = orig_get_platform


@contextlib.contextmanager
def scope_modules():
    orig_modules = frozenset(sys.modules)
    orig_path = list(sys.path)

    try:
        yield
    finally:
        for mod in frozenset(sys.modules).difference(orig_modules):
            del sys.modules[mod]
        sys.path = orig_path


def build_wheel_impl(args, wheel_dir: Path):
    build_sys = get_toml(args.pyproject_toml).get("build-system", {})
    backend_s = args.backend
    if backend_s is None:
        backend_s = build_sys.get("build-backend", args.fallback_backend)
        if backend_s is None:
            raise RuntimeError(
                "pyproject.toml is missing or does not specify build-backend "
                "and --no-fallback-backend specified")
    package, _, obj = backend_s.partition(":")

    zip_ctx = (contextlib.nullcontext if args.allow_compressed
               else disable_zip_compression)

    if args.sysroot is not None:
        if os.name == "nt":
            raise RuntimeError("--sysroot is not supported on Windows")
        sysconfig_ctx = patch_sysconfig(args.sysroot,
                                        args.prefix)
    else:
        sysconfig_ctx = contextlib.nullcontext()

    with zip_ctx(), sysconfig_ctx, scope_modules():
        def safe_samefile(path, cwd):
            try:
                return cwd.samefile(path)
            except Exception:
                return False

        # strip the current directory from sys.path
        cwd = Path.cwd()
        sys.path = [x for x in sys.path if not safe_samefile(x, cwd)]
        sys.path[:0] = build_sys.get("backend-path", [])
        backend = importlib.import_module(package)

        if obj:
            for name in obj.split("."):
                backend = getattr(backend, name)

        wheel_dir.mkdir(parents=True, exist_ok=True)

        logger.info(f"Building wheel via backend {backend_s}")
        wheel_name = backend.build_wheel(str(wheel_dir), args.config_json)
        logger.info(f"The backend produced {wheel_dir / wheel_name}")

        return wheel_name



import pluggy

hookspec = pluggy.HookspecMarker("hatch")


@hookspec
def hatch_register_version_source() -> None:
    """Register new classes that adhere to the version source interface."""


@hookspec
def hatch_register_builder() -> None:
    """Register new classes that adhere to the builder interface."""


@hookspec
def hatch_register_build_hook() -> None:
    """Register new classes that adhere to the build hook interface."""


@hookspec
def hatch_register_metadata_hook() -> None:
    """Register new classes that adhere to the metadata hook interface."""



# Copyright 2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import json
import os

import portage


class PurgeRevisions:
    short_desc = "Purge repo_revisions history file."

    @staticmethod
    def name():
        return "revisions"

    def __init__(self, settings=None):
        """Class init function

        @param settings: optional portage.config instance to get EROOT from.
        """
        self._settings = settings

    @property
    def settings(self):
        return self._settings or portage.settings

    def purgeallrepos(self, **kwargs):
        """Purge revisions for all repos"""
        repo_revisions_file = os.path.join(
            self.settings["EROOT"], portage.const.REPO_REVISIONS
        )
        msgs = []
        try:
            os.stat(repo_revisions_file)
        except FileNotFoundError:
            pass
        except OSError as e:
            msgs.append(f"{repo_revisions_file}: {e}")
        else:
            repo_revisions_lock = None
            try:
                repo_revisions_lock = portage.locks.lockfile(repo_revisions_file)
                os.unlink(repo_revisions_file)
            except FileNotFoundError:
                pass
            except OSError as e:
                msgs.append(f"{repo_revisions_file}: {e}")
            finally:
                if repo_revisions_lock is not None:
                    portage.locks.unlockfile(repo_revisions_lock)
        return (not msgs, msgs)

    def purgerepos(self, **kwargs):
        """Purge revisions for specified repos"""
        options = kwargs.get("options", None)
        if options:
            repo_names = options.get("purgerepos", "")
        if isinstance(repo_names, str):
            repo_names = repo_names.split()

        repo_revisions_file = os.path.join(
            self.settings["EROOT"], portage.const.REPO_REVISIONS
        )
        msgs = []
        try:
            os.stat(repo_revisions_file)
        except FileNotFoundError:
            pass
        except OSError as e:
            msgs.append(f"{repo_revisions_file}: {e}")
        else:
            repo_revisions_lock = None
            try:
                repo_revisions_lock = portage.locks.lockfile(repo_revisions_file)
                with open(repo_revisions_file, encoding="utf8") as f:
                    if os.fstat(f.fileno()).st_size:
                        previous_revisions = json.load(f)
                repo_revisions = (
                    {} if previous_revisions is None else previous_revisions.copy()
                )
                for repo_name in repo_names:
                    repo_revisions.pop(repo_name, None)
                if not repo_revisions:
                    os.unlink(repo_revisions_file)
                elif repo_revisions != previous_revisions:
                    f = portage.util.atomic_ofstream(repo_revisions_file)
                    json.dump(repo_revisions, f, ensure_ascii=False, sort_keys=True)
                    f.close()
            except OSError as e:
                msgs.append(f"{repo_revisions_file}: {e}")
            finally:
                if repo_revisions_lock is not None:
                    portage.locks.unlockfile(repo_revisions_lock)
        return (not msgs, msgs)



# Copyright 2012-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class TestDepend(TestCase):
    ebuilds = {
        "dev-libs/A-1": {
            "IUSE": "test",
            "DEPEND": "test? ( dev-libs/B )",
        },
        "dev-libs/B-1": {},
    }

    installed = {
        "dev-libs/A-1": {
            "USE": "",
            "IUSE": "test",
            "DEPEND": "test? ( dev-libs/B )",
        },
    }

    def test_default_use_test(self):
        """
        Test that FEATURES=test enables USE=test by default.
        """
        user_config = {"make.conf": ("FEATURES=test", 'USE=""')}
        test_case = ResolverPlaygroundTestCase(
            ["dev-libs/A"],
            options={},
            success=True,
            mergelist=["dev-libs/B-1", "dev-libs/A-1"],
        )

        playground = ResolverPlayground(
            ebuilds=self.ebuilds, user_config=user_config, debug=False
        )
        try:
            playground.run_TestCase(test_case)
            self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def test_no_forced_use_test(self):
        """
        Test that FEATURES=test no longer forces USE=test.
        """
        user_config = {"make.conf": ("FEATURES=test", 'USE="-test"')}
        test_case = ResolverPlaygroundTestCase(
            ["dev-libs/A"], options={}, success=True, mergelist=["dev-libs/A-1"]
        )

        playground = ResolverPlayground(
            ebuilds=self.ebuilds, user_config=user_config, debug=False
        )
        try:
            playground.run_TestCase(test_case)
            self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def test_newuse(self):
        """
        Test that --newuse now detects USE=test changes.
        """
        user_config = {"make.conf": ("FEATURES=test", 'USE=""')}
        test_case = ResolverPlaygroundTestCase(
            ["dev-libs/A"],
            options={"--newuse": True, "--selective": True},
            success=True,
            mergelist=["dev-libs/B-1", "dev-libs/A-1"],
        )

        playground = ResolverPlayground(
            ebuilds=self.ebuilds, user_config=user_config, debug=False
        )
        try:
            playground.run_TestCase(test_case)
            self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2021-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["bin_entry_point"]

import sys

from portage.const import PORTAGE_BIN_PATH
from portage import os


def bin_entry_point():
    """
    Adjust sys.argv[0] to point to a script in PORTAGE_BIN_PATH, and
    then execute the script, in order to implement entry_points when
    portage has been installed by pip.
    """
    script_path = os.path.join(PORTAGE_BIN_PATH, os.path.basename(sys.argv[0]))
    if os.access(script_path, os.X_OK):
        sys.argv = [
            sys.executable,
            script_path,
        ] + sys.argv[1:]
        os.execvp(sys.argv[0], sys.argv)
    else:
        print("File not found:", script_path, file=sys.stderr)
        return 127



# Copyright 1999-2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.elog import mod_echo


def _flush_elog_mod_echo():
    """
    Dump the mod_echo output now so that our other
    notifications are shown last.
    @rtype: bool
    @return: True if messages were shown, False otherwise.
    """
    messages_shown = bool(mod_echo._items)
    mod_echo.finalize()
    return messages_shown



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import Union

from setuptools import Distribution

from .. import _log

if TYPE_CHECKING:
    from .pyproject_reading import PyProjectData

log = _log.log.getChild("version_inference")


@dataclass
class VersionInferenceConfig:
    """Configuration for version inference."""

    dist_name: str | None
    pyproject_data: PyProjectData | None
    overrides: dict[str, Any] | None

    def apply(self, dist: Distribution) -> None:
        """Apply version inference to the distribution."""
        version_string = infer_version_string(
            self.dist_name,
            self.pyproject_data,  # type: ignore[arg-type]
            self.overrides,
            force_write_version_files=True,
        )
        dist.metadata.version = version_string

        # Mark that this version was set by infer_version if overrides is None (infer_version context)
        if self.overrides is None:
            dist._setuptools_scm_version_set_by_infer = True  # type: ignore[attr-defined]


@dataclass
class VersionInferenceWarning:
    """Error message for user."""

    message: str

    def apply(self, dist: Distribution) -> None:
        """Apply error handling to the distribution."""
        import warnings

        warnings.warn(self.message)


@dataclass(frozen=True)
class VersionInferenceNoOp:
    """No operation result - silent skip."""

    def apply(self, dist: Distribution) -> None:
        """Apply no-op to the distribution."""


VersionInferenceResult = Union[
    VersionInferenceConfig,  # Proceed with inference
    VersionInferenceWarning,  # Show warning
    VersionInferenceNoOp,  # Don't infer (silent)
]


def infer_version_string(
    dist_name: str | None,
    pyproject_data: PyProjectData,
    overrides: dict[str, Any] | None = None,
    *,
    force_write_version_files: bool = False,
) -> str:
    """
    Compute the inferred version string from the given inputs without requiring a
    setuptools Distribution instance. This is a pure helper that simplifies
    integration tests by avoiding file I/O and side effects on a Distribution.

    Parameters:
        dist_name: Optional distribution name (used for overrides and env scoping)
        pyproject_data: Parsed PyProjectData (may be constructed via for_testing())
        overrides: Optional override configuration (same keys as [tool.setuptools_scm])
        force_write_version_files: When True, apply write_to/version_file effects

    Returns:
        The computed version string.
    """
    from .. import _config as _config_module
    from .._get_version_impl import _get_version
    from .._get_version_impl import _version_missing

    config = _config_module.Configuration.from_file(
        dist_name=dist_name, pyproject_data=pyproject_data, **(overrides or {})
    )

    maybe_version = _get_version(
        config, force_write_version_files=force_write_version_files
    )
    if maybe_version is None:
        _version_missing(config)
    return maybe_version


def get_version_inference_config(
    dist_name: str | None,
    current_version: str | None,
    pyproject_data: PyProjectData,
    overrides: dict[str, Any] | None = None,
) -> VersionInferenceResult:
    """
    Determine whether and how to perform version inference.

    Args:
        dist_name: The distribution name
        current_version: Current version if any
        pyproject_data: PyProjectData from parser (None if file doesn't exist)
        overrides: Override configuration (None for no overrides)

    Returns:
        VersionInferenceResult with the decision and configuration
    """

    config = VersionInferenceConfig(
        dist_name=dist_name,
        pyproject_data=pyproject_data,
        overrides=overrides,
    )

    inference_implied = pyproject_data.should_infer() or overrides is not None

    if inference_implied:
        if current_version is None:
            return config
        else:
            return VersionInferenceWarning(
                f"version of {dist_name} already set",
            )
    else:
        return VersionInferenceNoOp()



import sys

# Explicitly use the ``"locale"`` encoding in versions that support it,
# otherwise just rely on the implicit handling of ``encoding=None``.
# Since all platforms that support ``EncodingWarning`` also support
# ``encoding="locale"``, this can be used to suppress the warning.
# However, please try to use UTF-8 when possible
# (.pth files are the notorious exception: python/cpython#77102, pypa/setuptools#3937).
LOCALE_ENCODING = "locale" if sys.version_info >= (3, 10) else None



# Copyright 2013-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import logging
import time
import threading

import portage
from portage import os
from portage.util._async.TaskScheduler import TaskScheduler
from _emerge.CompositeTask import CompositeTask
from .FetchIterator import FetchIterator
from .DeletionIterator import DeletionIterator

logger = logging.getLogger(__name__)


class MirrorDistTask(CompositeTask):
    __slots__ = ("_config", "_fetch_iterator", "_term_rlock", "_term_callback_handle")

    def __init__(self, config):
        CompositeTask.__init__(self, scheduler=config.event_loop)
        self._config = config
        self._term_rlock = threading.RLock()
        self._term_callback_handle = None
        self._fetch_iterator = None

    def _start(self):
        self._fetch_iterator = FetchIterator(self._config)
        fetch = TaskScheduler(
            iter(self._fetch_iterator),
            max_jobs=self._config.options.jobs,
            max_load=self._config.options.load_average,
            event_loop=self._config.event_loop,
        )
        self._start_task(fetch, self._fetch_exit)

    def _fetch_exit(self, fetch):
        self._assert_current(fetch)
        if self._was_cancelled():
            self._async_wait()
            return

        if self._config.options.delete:
            deletion = TaskScheduler(
                iter(DeletionIterator(self._config)),
                max_jobs=self._config.options.jobs,
                max_load=self._config.options.load_average,
                event_loop=self._config.event_loop,
            )
            self._start_task(deletion, self._deletion_exit)
            return

        self._post_deletion()

    def _deletion_exit(self, deletion):
        self._assert_current(deletion)
        if self._was_cancelled():
            self._async_wait()
            return

        self._post_deletion()

    def _post_deletion(self):
        if self._config.options.recycle_db is not None:
            self._update_recycle_db()

        if self._config.options.scheduled_deletion_log is not None:
            self._scheduled_deletion_log()

        self._summary()

        self.returncode = os.EX_OK
        self._current_task = None
        self._async_wait()

    def _update_recycle_db(self):
        start_time = self._config.start_time
        recycle_dir = self._config.options.recycle_dir
        recycle_db = self._config.recycle_db
        r_deletion_delay = self._config.options.recycle_deletion_delay

        # Use a dict optimize access.
        recycle_db_cache = dict(recycle_db.items())

        for filename in os.listdir(recycle_dir):
            recycle_file = os.path.join(recycle_dir, filename)

            try:
                st = os.stat(recycle_file)
            except OSError as e:
                if e.errno not in (errno.ENOENT, errno.ESTALE):
                    logger.error(f"stat failed for '{filename}' in recycle: {e}")
                continue

            value = recycle_db_cache.pop(filename, None)
            if value is None:
                logger.debug(f"add '{filename}' to recycle db")
                recycle_db[filename] = (st.st_size, start_time)
            else:
                r_size, r_time = value
                if int(r_size) != st.st_size:
                    recycle_db[filename] = (st.st_size, start_time)
                elif r_time + r_deletion_delay < start_time:
                    if self._config.options.dry_run:
                        logger.info(f"dry-run: delete '{filename}' from recycle")
                        logger.info(f"drop '{filename}' from recycle db")
                    else:
                        try:
                            os.unlink(recycle_file)
                        except OSError as e:
                            if e.errno not in (errno.ENOENT, errno.ESTALE):
                                logger.error(
                                    f"delete '{filename}' from recycle failed: {e}"
                                )
                        else:
                            logger.debug(f"delete '{filename}' from recycle")
                            try:
                                del recycle_db[filename]
                            except KeyError:
                                pass
                            else:
                                logger.debug(f"drop '{filename}' from recycle db")

        # Existing files were popped from recycle_db_cache,
        # so any remaining entries are for files that no
        # longer exist.
        for filename in recycle_db_cache:
            try:
                del recycle_db[filename]
            except KeyError:
                pass
            else:
                logger.debug(f"drop non-existent '{filename}' from recycle db")

    def _scheduled_deletion_log(self):
        start_time = self._config.start_time
        dry_run = self._config.options.dry_run
        deletion_delay = self._config.options.deletion_delay
        distfiles_db = self._config.distfiles_db

        date_map = {}
        for filename, timestamp in self._config.deletion_db.items():
            date = timestamp + deletion_delay
            if date < start_time:
                date = start_time
            date = time.strftime("%Y-%m-%d", time.gmtime(date))
            date_files = date_map.get(date)
            if date_files is None:
                date_files = []
                date_map[date] = date_files
            date_files.append(filename)

        if dry_run:
            logger.warning(
                "dry-run: scheduled-deletions log "
                "will be summarized via logging.info"
            )

        lines = []
        for date in sorted(date_map):
            date_files = date_map[date]
            if dry_run:
                logger.info(
                    f"dry-run: scheduled deletions for {date}: {len(date_files)} files"
                )
            lines.append(f"{date}\n")
            for filename in date_files:
                cpv = "unknown"
                if distfiles_db is not None:
                    cpv = distfiles_db.get(filename, cpv)
                lines.append(f"\t{filename}\t{cpv}\n")

        if not dry_run:
            portage.util.write_atomic(
                self._config.options.scheduled_deletion_log, "".join(lines)
            )

    def _summary(self):
        elapsed_time = time.time() - self._config.start_time
        fail_count = len(self._config.file_failures)
        delete_count = self._config.delete_count
        scheduled_deletion_count = self._config.scheduled_deletion_count - delete_count
        added_file_count = self._config.added_file_count
        added_byte_count = self._config.added_byte_count

        logger.info("finished in %i seconds" % elapsed_time)
        logger.info("failed to fetch %i files" % fail_count)
        logger.info("deleted %i files" % delete_count)
        logger.info("deletion of %i files scheduled" % scheduled_deletion_count)
        logger.info("added %i files" % added_file_count)
        logger.info("added %i bytes total" % added_byte_count)

    def _cleanup(self):
        """
        Cleanup any callbacks that have been registered with the global
        event loop.
        """
        # The self._term_callback_handle attribute requires locking
        # since it's modified by the thread safe terminate method.
        with self._term_rlock:
            if self._term_callback_handle not in (None, False):
                self._term_callback_handle.cancel()
            # This prevents the terminate method from scheduling
            # any more callbacks (since _cleanup must eliminate all
            # callbacks in order to ensure complete cleanup).
            self._term_callback_handle = False

    def terminate(self):
        with self._term_rlock:
            if self._term_callback_handle is None:
                self._term_callback_handle = self.scheduler.call_soon_threadsafe(
                    self._term_callback
                )

    def _term_callback(self):
        if self._fetch_iterator is not None:
            self._fetch_iterator.terminate()
        self.cancel()
        if self.returncode is None:
            # In this case, the exit callback for self._current_task will
            # trigger notification of exit listeners. Don't call _async_wait()
            # yet, since that could trigger event loop recursion if the
            # current (cancelled) task's exit callback does not set the
            # returncode first.
            pass
        else:
            self._async_wait()

    def _async_wait(self):
        """
        Override _async_wait to call self._cleanup().
        """
        self._cleanup()
        super()._async_wait()



import re
from configparser import ConfigParser
from inspect import cleandoc

import jaraco.path
import pytest
import tomli_w
from path import Path

import setuptools  # noqa: F401 # force distutils.core to be patched
from setuptools.config.pyprojecttoml import (
    _ToolsTypoInMetadata,
    apply_configuration,
    expand_configuration,
    read_configuration,
    validate,
)
from setuptools.dist import Distribution
from setuptools.errors import OptionError

import distutils.core

EXAMPLE = """
[project]
name = "myproj"
keywords = ["some", "key", "words"]
dynamic = ["version", "readme"]
requires-python = ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*"
dependencies = [
    'importlib-metadata>=0.12;python_version<"3.8"',
    'importlib-resources>=1.0;python_version<"3.7"',
    'pathlib2>=2.3.3,<3;python_version < "3.4" and sys.platform != "win32"',
]

[project.optional-dependencies]
docs = [
    "sphinx>=3",
    "sphinx-argparse>=0.2.5",
    "sphinx-rtd-theme>=0.4.3",
]
testing = [
    "pytest>=1",
    "coverage>=3,<5",
]

[project.scripts]
exec = "pkg.__main__:exec"

[build-system]
requires = ["setuptools", "wheel"]
build-backend = "setuptools.build_meta"

[tool.setuptools]
package-dir = {"" = "src"}
zip-safe = true
platforms = ["any"]

[tool.setuptools.packages.find]
where = ["src"]

[tool.setuptools.cmdclass]
sdist = "pkg.mod.CustomSdist"

[tool.setuptools.dynamic.version]
attr = "pkg.__version__.VERSION"

[tool.setuptools.dynamic.readme]
file = ["README.md"]
content-type = "text/markdown"

[tool.setuptools.package-data]
"*" = ["*.txt"]

[tool.setuptools.data-files]
"data" = ["_files/*.txt"]

[tool.distutils.sdist]
formats = "gztar"

[tool.distutils.bdist_wheel]
universal = true
"""


def create_example(path, pkg_root):
    files = {
        "pyproject.toml": EXAMPLE,
        "README.md": "hello world",
        "_files": {
            "file.txt": "",
        },
    }
    packages = {
        "pkg": {
            "__init__.py": "",
            "mod.py": "class CustomSdist: pass",
            "__version__.py": "VERSION = (3, 10)",
            "__main__.py": "def exec(): print('hello')",
        },
    }

    assert pkg_root  # Meta-test: cannot be empty string.

    if pkg_root == ".":
        files = {**files, **packages}
        # skip other files: flat-layout will raise error for multi-package dist
    else:
        # Use this opportunity to ensure namespaces are discovered
        files[pkg_root] = {**packages, "other": {"nested": {"__init__.py": ""}}}

    jaraco.path.build(files, prefix=path)


def verify_example(config, path, pkg_root):
    pyproject = path / "pyproject.toml"
    pyproject.write_text(tomli_w.dumps(config), encoding="utf-8")
    expanded = expand_configuration(config, path)
    expanded_project = expanded["project"]
    assert read_configuration(pyproject, expand=True) == expanded
    assert expanded_project["version"] == "3.10"
    assert expanded_project["readme"]["text"] == "hello world"
    assert "packages" in expanded["tool"]["setuptools"]
    if pkg_root == ".":
        # Auto-discovery will raise error for multi-package dist
        assert set(expanded["tool"]["setuptools"]["packages"]) == {"pkg"}
    else:
        assert set(expanded["tool"]["setuptools"]["packages"]) == {
            "pkg",
            "other",
            "other.nested",
        }
    assert expanded["tool"]["setuptools"]["include-package-data"] is True
    assert "" in expanded["tool"]["setuptools"]["package-data"]
    assert "*" not in expanded["tool"]["setuptools"]["package-data"]
    assert expanded["tool"]["setuptools"]["data-files"] == [
        ("data", ["_files/file.txt"])
    ]


def test_read_configuration(tmp_path):
    create_example(tmp_path, "src")
    pyproject = tmp_path / "pyproject.toml"

    config = read_configuration(pyproject, expand=False)
    assert config["project"].get("version") is None
    assert config["project"].get("readme") is None

    verify_example(config, tmp_path, "src")


@pytest.mark.parametrize(
    ("pkg_root", "opts"),
    [
        (".", {}),
        ("src", {}),
        ("lib", {"packages": {"find": {"where": ["lib"]}}}),
    ],
)
def test_discovered_package_dir_with_attr_directive_in_config(tmp_path, pkg_root, opts):
    create_example(tmp_path, pkg_root)

    pyproject = tmp_path / "pyproject.toml"

    config = read_configuration(pyproject, expand=False)
    assert config["project"].get("version") is None
    assert config["project"].get("readme") is None
    config["tool"]["setuptools"].pop("packages", None)
    config["tool"]["setuptools"].pop("package-dir", None)

    config["tool"]["setuptools"].update(opts)
    verify_example(config, tmp_path, pkg_root)


ENTRY_POINTS = {
    "console_scripts": {"a": "mod.a:func"},
    "gui_scripts": {"b": "mod.b:func"},
    "other": {"c": "mod.c:func [extra]"},
}


class TestEntryPoints:
    def write_entry_points(self, tmp_path):
        entry_points = ConfigParser()
        entry_points.read_dict(ENTRY_POINTS)
        with open(tmp_path / "entry-points.txt", "w", encoding="utf-8") as f:
            entry_points.write(f)

    def pyproject(self, dynamic=None):
        project = {"dynamic": dynamic or ["scripts", "gui-scripts", "entry-points"]}
        tool = {"dynamic": {"entry-points": {"file": "entry-points.txt"}}}
        return {"project": project, "tool": {"setuptools": tool}}

    def test_all_listed_in_dynamic(self, tmp_path):
        self.write_entry_points(tmp_path)
        expanded = expand_configuration(self.pyproject(), tmp_path)
        expanded_project = expanded["project"]
        assert len(expanded_project["scripts"]) == 1
        assert expanded_project["scripts"]["a"] == "mod.a:func"
        assert len(expanded_project["gui-scripts"]) == 1
        assert expanded_project["gui-scripts"]["b"] == "mod.b:func"
        assert len(expanded_project["entry-points"]) == 1
        assert expanded_project["entry-points"]["other"]["c"] == "mod.c:func [extra]"

    @pytest.mark.parametrize("missing_dynamic", ("scripts", "gui-scripts"))
    def test_scripts_not_listed_in_dynamic(self, tmp_path, missing_dynamic):
        self.write_entry_points(tmp_path)
        dynamic = {"scripts", "gui-scripts", "entry-points"} - {missing_dynamic}

        msg = f"defined outside of `pyproject.toml`:.*{missing_dynamic}"
        with pytest.raises(OptionError, match=re.compile(msg, re.DOTALL)):
            expand_configuration(self.pyproject(dynamic), tmp_path)


class TestClassifiers:
    def test_dynamic(self, tmp_path):
        # Let's create a project example that has dynamic classifiers
        # coming from a txt file.
        create_example(tmp_path, "src")
        classifiers = cleandoc(
            """
            Framework :: Flask
            Programming Language :: Haskell
            """
        )
        (tmp_path / "classifiers.txt").write_text(classifiers, encoding="utf-8")

        pyproject = tmp_path / "pyproject.toml"
        config = read_configuration(pyproject, expand=False)
        dynamic = config["project"]["dynamic"]
        config["project"]["dynamic"] = list({*dynamic, "classifiers"})
        dynamic_config = config["tool"]["setuptools"]["dynamic"]
        dynamic_config["classifiers"] = {"file": "classifiers.txt"}

        # When the configuration is expanded,
        # each line of the file should be an different classifier.
        validate(config, pyproject)
        expanded = expand_configuration(config, tmp_path)

        assert set(expanded["project"]["classifiers"]) == {
            "Framework :: Flask",
            "Programming Language :: Haskell",
        }

    def test_dynamic_without_config(self, tmp_path):
        config = """
        [project]
        name = "myproj"
        version = '42'
        dynamic = ["classifiers"]
        """

        pyproject = tmp_path / "pyproject.toml"
        pyproject.write_text(cleandoc(config), encoding="utf-8")
        with pytest.raises(OptionError, match="No configuration .* .classifiers."):
            read_configuration(pyproject)

    def test_dynamic_readme_from_setup_script_args(self, tmp_path):
        config = """
        [project]
        name = "myproj"
        version = '42'
        dynamic = ["readme"]
        """
        pyproject = tmp_path / "pyproject.toml"
        pyproject.write_text(cleandoc(config), encoding="utf-8")
        dist = Distribution(attrs={"long_description": "42"})
        # No error should occur because of missing `readme`
        dist = apply_configuration(dist, pyproject)
        assert dist.metadata.long_description == "42"

    def test_dynamic_without_file(self, tmp_path):
        config = """
        [project]
        name = "myproj"
        version = '42'
        dynamic = ["classifiers"]

        [tool.setuptools.dynamic]
        classifiers = {file = ["classifiers.txt"]}
        """

        pyproject = tmp_path / "pyproject.toml"
        pyproject.write_text(cleandoc(config), encoding="utf-8")
        with pytest.warns(UserWarning, match="File .*classifiers.txt. cannot be found"):
            expanded = read_configuration(pyproject)
        assert "classifiers" not in expanded["project"]


class TestImportNames:
    EXAMPLES = [
        'import-names = ["hello", "world"]',
        'import-namespaces = ["hello", "world"]',
        'dynamic = ["import-names"]',
        'dynamic = ["import-namespaces"]',
    ]

    @pytest.mark.parametrize("example", EXAMPLES)
    def test_not_implemented(self, monkeypatch, tmp_path, example):
        monkeypatch.chdir(tmp_path)
        pyproject = Path("pyproject.toml")
        toml_config = f"""
        [project]
        name = 'proj'
        version = '42'
        {example}
        """
        pyproject.write_text(cleandoc(toml_config), encoding="utf-8")
        with pytest.raises(NotImplementedError, match='import-names'):
            apply_configuration(Distribution({}), pyproject)


@pytest.mark.parametrize(
    "example",
    (
        """
        [project]
        name = "myproj"
        version = "1.2"

        [my-tool.that-disrespect.pep518]
        value = 42
        """,
    ),
)
def test_ignore_unrelated_config(tmp_path, example):
    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(cleandoc(example), encoding="utf-8")

    # Make sure no error is raised due to 3rd party configs in pyproject.toml
    assert read_configuration(pyproject) is not None


@pytest.mark.parametrize(
    ("example", "error_msg"),
    [
        (
            """
            [project]
            name = "myproj"
            version = "1.2"
            requires = ['pywin32; platform_system=="Windows"' ]
            """,
            "configuration error: .project. must not contain ..requires.. properties",
        ),
    ],
)
def test_invalid_example(tmp_path, example, error_msg):
    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(cleandoc(example), encoding="utf-8")

    pattern = re.compile(
        f"invalid pyproject.toml.*{error_msg}.*", re.MULTILINE | re.DOTALL
    )
    with pytest.raises(ValueError, match=pattern):
        read_configuration(pyproject)


@pytest.mark.parametrize("config", ("", "[tool.something]\nvalue = 42"))
def test_empty(tmp_path, config):
    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(config, encoding="utf-8")

    # Make sure no error is raised
    assert read_configuration(pyproject) == {}


@pytest.mark.parametrize("config", ("[project]\nname = 'myproj'\nversion='42'\n",))
def test_include_package_data_by_default(tmp_path, config):
    """Builds with ``pyproject.toml`` should consider ``include-package-data=True`` as
    default.
    """
    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(config, encoding="utf-8")

    config = read_configuration(pyproject)
    assert config["tool"]["setuptools"]["include-package-data"] is True


def test_include_package_data_in_setuppy(tmp_path):
    """Builds with ``pyproject.toml`` should consider ``include_package_data`` set in
    ``setup.py``.

    See https://github.com/pypa/setuptools/issues/3197#issuecomment-1079023889
    """
    files = {
        "pyproject.toml": "[project]\nname = 'myproj'\nversion='42'\n",
        "setup.py": "__import__('setuptools').setup(include_package_data=False)",
    }
    jaraco.path.build(files, prefix=tmp_path)

    with Path(tmp_path):
        dist = distutils.core.run_setup("setup.py", {}, stop_after="config")

    assert dist.get_name() == "myproj"
    assert dist.get_version() == "42"
    assert dist.include_package_data is False


def test_warn_tools_typo(tmp_path):
    """Test that the common ``tools.setuptools`` typo in ``pyproject.toml`` issues a warning

    See https://github.com/pypa/setuptools/issues/4150
    """
    config = """
    [build-system]
    requires = ["setuptools"]
    build-backend = "setuptools.build_meta"

    [project]
    name = "myproj"
    version = '42'

    [tools.setuptools]
    packages = ["package"]
    """

    pyproject = tmp_path / "pyproject.toml"
    pyproject.write_text(cleandoc(config), encoding="utf-8")

    with pytest.warns(_ToolsTypoInMetadata):
        read_configuration(pyproject)



from __future__ import annotations

import argparse
import json
import os
import sys

from pathlib import Path
from typing import Any

from setuptools_scm import Configuration
from setuptools_scm._file_finders import find_files
from setuptools_scm._get_version_impl import _get_version
from setuptools_scm.discover import walk_potential_roots


def main(args: list[str] | None = None) -> int:
    opts = _get_cli_opts(args)
    inferred_root: str = opts.root or "."

    pyproject = opts.config or _find_pyproject(inferred_root)

    try:
        config = Configuration.from_file(
            pyproject,
            root=(os.path.abspath(opts.root) if opts.root is not None else None),
        )
    except (LookupError, FileNotFoundError) as ex:
        # no pyproject.toml OR no [tool.setuptools_scm]
        print(
            f"Warning: could not use {os.path.relpath(pyproject)},"
            " using default configuration.\n"
            f" Reason: {ex}.",
            file=sys.stderr,
        )
        config = Configuration(root=inferred_root)
    version: str | None
    if opts.no_version:
        version = "0.0.0+no-version-was-requested.fake-version"
    else:
        version = _get_version(
            config, force_write_version_files=opts.force_write_version_files
        )
    if version is None:
        raise SystemExit("ERROR: no version found for", opts)
    if opts.strip_dev:
        version = version.partition(".dev")[0]

    return command(opts, version, config)


def _get_cli_opts(args: list[str] | None) -> argparse.Namespace:
    prog = "python -m setuptools_scm"
    desc = "Print project version according to SCM metadata"
    parser = argparse.ArgumentParser(prog, description=desc)
    # By default, help for `--help` starts with lower case, so we keep the pattern:
    parser.add_argument(
        "-r",
        "--root",
        default=None,
        help='directory managed by the SCM, default: inferred from config file, or "."',
    )
    parser.add_argument(
        "-c",
        "--config",
        default=None,
        metavar="PATH",
        help="path to 'pyproject.toml' with setuptools-scm config, "
        "default: looked up in the current or parent directories",
    )
    parser.add_argument(
        "--strip-dev",
        action="store_true",
        help="remove the dev/local parts of the version before printing the version",
    )
    parser.add_argument(
        "-N",
        "--no-version",
        action="store_true",
        help="do not include package version in the output",
    )
    output_formats = ["json", "plain", "key-value"]
    parser.add_argument(
        "-f",
        "--format",
        type=str.casefold,
        default="plain",
        help="specify output format",
        choices=output_formats,
    )
    parser.add_argument(
        "-q",
        "--query",
        type=str.casefold,
        nargs="*",
        help="display setuptools-scm settings according to query, "
        "e.g. dist_name, do not supply an argument in order to "
        "print a list of valid queries.",
    )
    parser.add_argument(
        "--force-write-version-files",
        action="store_true",
        help="trigger to write the content of the version files\n"
        "its recommended to use normal/editable installation instead)",
    )
    sub = parser.add_subparsers(title="extra commands", dest="command", metavar="")
    # We avoid `metavar` to prevent printing repetitive information
    desc = "List information about the package, e.g. included files"
    sub.add_parser("ls", help=desc[0].lower() + desc[1:], description=desc)

    # Add create-archival-file subcommand
    archival_desc = "Create .git_archival.txt file for git archive support"
    archival_parser = sub.add_parser(
        "create-archival-file",
        help=archival_desc[0].lower() + archival_desc[1:],
        description=archival_desc,
    )
    archival_group = archival_parser.add_mutually_exclusive_group(required=True)
    archival_group.add_argument(
        "--stable",
        action="store_true",
        help="create stable archival file (recommended, no branch names)",
    )
    archival_group.add_argument(
        "--full",
        action="store_true",
        help="create full archival file with branch information (can cause instability)",
    )
    archival_parser.add_argument(
        "--force", action="store_true", help="overwrite existing .git_archival.txt file"
    )
    return parser.parse_args(args)


# flake8: noqa: C901
def command(opts: argparse.Namespace, version: str, config: Configuration) -> int:
    data: dict[str, Any] = {}

    if opts.command == "ls":
        opts.query = ["files"]

    if opts.command == "create-archival-file":
        return _create_archival_file(opts, config)

    if opts.query == []:
        opts.no_version = True
        sys.stderr.write("Available queries:\n\n")
        opts.query = ["queries"]
        data["queries"] = ["files", *config.__dataclass_fields__]

    if opts.query is None:
        opts.query = []

    if not opts.no_version:
        data["version"] = version

    if "files" in opts.query:
        data["files"] = find_files(config.root)

    for q in opts.query:
        if q in ["files", "queries", "version"]:
            continue

        try:
            if q.startswith("_"):
                raise AttributeError()
            data[q] = getattr(config, q)
        except AttributeError:
            sys.stderr.write(f"Error: unknown query: '{q}'\n")
            return 1

    if opts.format == "json":
        print(json.dumps(data, indent=2))

    if opts.format == "plain":
        _print_plain(data)

    if opts.format == "key-value":
        _print_key_value(data)

    return 0


def _print_plain(data: dict[str, Any]) -> None:
    version = data.pop("version", None)
    if version:
        print(version)
    files = data.pop("files", [])
    for file_ in files:
        print(file_)
    queries = data.pop("queries", [])
    for query in queries:
        print(query)
    if data:
        print("\n".join(data.values()))


def _print_key_value(data: dict[str, Any]) -> None:
    for key, value in data.items():
        if isinstance(value, str):
            print(f"{key} = {value}")
        else:
            str_value = "\n  ".join(value)
            print(f"{key} = {str_value}")


def _find_pyproject(parent: str) -> str:
    for directory in walk_potential_roots(os.path.abspath(parent)):
        pyproject = os.path.join(directory, "pyproject.toml")
        if os.path.isfile(pyproject):
            return pyproject

    return os.path.abspath(
        "pyproject.toml"
    )  # use default name to trigger the default errors


def _create_archival_file(opts: argparse.Namespace, config: Configuration) -> int:
    """Create .git_archival.txt file with appropriate content."""
    archival_path = Path(config.root, ".git_archival.txt")

    # Check if file exists and force flag
    if archival_path.exists() and not opts.force:
        print(
            f"Error: {archival_path} already exists. Use --force to overwrite.",
            file=sys.stderr,
        )
        return 1

    if opts.stable:
        content = _get_stable_archival_content()
        print("Creating stable .git_archival.txt (recommended for releases)")
    elif opts.full:
        content = _get_full_archival_content()
        print("Creating full .git_archival.txt with branch information")
        print("WARNING: This can cause archive checksums to be unstable!")

    try:
        archival_path.write_text(content, encoding="utf-8")
        print(f"Created: {archival_path}")

        gitattributes_path = Path(config.root, ".gitattributes")
        needs_gitattributes = True

        if gitattributes_path.exists():
            # TODO: more nuanced check later
            gitattributes_content = gitattributes_path.read_text("utf-8")
            if (
                ".git_archival.txt" in gitattributes_content
                and "export-subst" in gitattributes_content
            ):
                needs_gitattributes = False

        if needs_gitattributes:
            print("\nNext steps:")
            print("1. Add this line to .gitattributes:")
            print("   .git_archival.txt  export-subst")
            print("2. Commit both files:")
            print("   git add .git_archival.txt .gitattributes")
            print("   git commit -m 'add git archive support'")
        else:
            print("\nNext step:")
            print("Commit the archival file:")
            print("   git add .git_archival.txt")
            print("   git commit -m 'update git archival file'")

        return 0
    except OSError as e:
        print(f"Error: Could not create {archival_path}: {e}", file=sys.stderr)
        return 1


def _get_stable_archival_content() -> str:
    """Generate stable archival file content (no branch names)."""
    return """\
node: $Format:%H$
node-date: $Format:%cI$
describe-name: $Format:%(describe:tags=true,match=*[0-9]*)$
"""


def _get_full_archival_content() -> str:
    """Generate full archival file content with branch information."""
    return """\
# WARNING: Including ref-names can make archive checksums unstable
# after commits are added post-release. Use only if describe-name is insufficient.
node: $Format:%H$
node-date: $Format:%cI$
describe-name: $Format:%(describe:tags=true,match=*[0-9]*)$
ref-names: $Format:%D$
"""



# Copyright 2018-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = (
    "ALL_COMPLETED",
    "FIRST_COMPLETED",
    "FIRST_EXCEPTION",
    "ensure_future",
    "CancelledError",
    "Future",
    "InvalidStateError",
    "Lock",
    "TimeoutError",
    "get_child_watcher",
    "get_event_loop",
    "set_child_watcher",
    "get_event_loop_policy",
    "set_event_loop_policy",
    "run",
    "shield",
    "sleep",
    "wait",
    "wait_for",
)

import sys
import types
import warnings
import weakref

import asyncio as _real_asyncio

# pylint: disable=redefined-builtin
from asyncio import (
    ALL_COMPLETED,
    CancelledError,
    FIRST_COMPLETED,
    FIRST_EXCEPTION,
    Future,
    InvalidStateError,
    Lock as _Lock,
    shield,
    TimeoutError,
    wait_for,
)

from inspect import iscoroutinefunction
import threading
from typing import Optional

import portage
from portage.process import atexit_register
from portage.util._eventloop.asyncio_event_loop import (
    AsyncioEventLoop as _AsyncioEventLoop,
)


_lock = threading.Lock()
_policy = None


def get_event_loop_policy():
    """
    Get the current event loop policy.

    @rtype: asyncio.AbstractEventLoopPolicy (or compatible)
    @return: the current event loop policy
    """
    from portage.util.futures.unix_events import _PortageEventLoopPolicy

    global _lock, _policy
    with _lock:
        if _policy is None:
            _policy = _PortageEventLoopPolicy()
        return _policy


def set_event_loop_policy(policy):
    """
    Set the current event loop policy. If policy is None, the default
    policy is restored.

    @type policy: asyncio.AbstractEventLoopPolicy or None
    @param policy: new event loop policy
    """
    from portage.util.futures.unix_events import _PortageEventLoopPolicy

    global _lock, _policy
    with _lock:
        _policy = policy or _PortageEventLoopPolicy()


def get_event_loop():
    """
    Equivalent to calling get_event_loop_policy().get_event_loop().

    @rtype: asyncio.AbstractEventLoop (or compatible)
    @return: the event loop for the current context
    """
    return get_event_loop_policy().get_event_loop()


def get_child_watcher():
    """Equivalent to calling get_event_loop_policy().get_child_watcher()."""
    return get_event_loop_policy().get_child_watcher()


def set_child_watcher(watcher):
    """Equivalent to calling
    get_event_loop_policy().set_child_watcher(watcher)."""
    return get_event_loop_policy().set_child_watcher(watcher)


# Emulate run since it's the preferred python API.
def run(coro):
    return _safe_loop().run_until_complete(coro)


run.__doc__ = _real_asyncio.run.__doc__


def create_subprocess_exec(*args, loop=None, **kwargs):
    """
    Create a subprocess.

    @param args: program and arguments
    @type args: str
    @param stdin: stdin file descriptor
    @type stdin: file or int
    @param stdout: stdout file descriptor
    @type stdout: file or int
    @param stderr: stderr file descriptor
    @type stderr: file or int
    @param close_fds: close file descriptors
    @type close_fds: bool
    @param loop: asyncio.AbstractEventLoop (or compatible)
    @type loop: event loop
    @type kwargs: varies
    @param kwargs: subprocess.Popen parameters
    @rtype: asyncio.subprocess.Process (or compatible)
    @return: asyncio.subprocess.Process interface
    """
    # Python 3.4 and later implement PEP 446, which makes newly
    # created file descriptors non-inheritable by default.
    kwargs.setdefault("close_fds", False)
    # Use the real asyncio create_subprocess_exec (loop argument
    # is deprecated since since Python 3.8).
    return ensure_future(
        _real_asyncio.create_subprocess_exec(*args, **kwargs), loop=loop
    )


def wait(futures, loop=None, timeout=None, return_when=ALL_COMPLETED):
    """
    Wraps asyncio.wait() and omits the loop argument which is not
    supported since python 3.10.
    """
    return _real_asyncio.wait(futures, timeout=timeout, return_when=return_when)


class Lock(_Lock):
    """
    Inject loop parameter for python3.9 or less in order to avoid
    "got Future <Future pending> attached to a different loop" errors.
    """

    def __init__(self, **kwargs):
        if sys.version_info >= (3, 10):
            kwargs.pop("loop", None)
        elif "loop" not in kwargs:
            kwargs["loop"] = _safe_loop()._loop
        super().__init__(**kwargs)


def ensure_future(coro_or_future, loop=None):
    """
    Wrap a coroutine or an awaitable in a future.

    If the argument is a Future, it is returned directly.

    @type coro_or_future: coroutine or Future
    @param coro_or_future: coroutine or future to wrap
    @type loop: asyncio.AbstractEventLoop (or compatible)
    @param loop: event loop
    @rtype: asyncio.Future (or compatible)
    @return: an instance of Future
    """
    if loop is None:
        return _real_asyncio.ensure_future(coro_or_future)

    loop = _wrap_loop(loop)
    if isinstance(loop._asyncio_wrapper, _AsyncioEventLoop):
        # Use the real asyncio loop and ensure_future.
        return _real_asyncio.ensure_future(
            coro_or_future, loop=loop._asyncio_wrapper._loop
        )

    if isinstance(coro_or_future, Future):
        return coro_or_future
    raise NotImplementedError


def sleep(delay, result=None, loop=None):
    """
    Create a future that completes after a given time (in seconds). If
    result is provided, it is produced to the caller when the future
    completes.

    @type delay: int or float
    @param delay: delay seconds
    @type result: object
    @param result: result of the future
    @type loop: asyncio.AbstractEventLoop (or compatible)
    @param loop: event loop
    @rtype: collections.abc.Coroutine or asyncio.Future
    @return: an instance of Coroutine or Future
    """
    if loop is None:
        return _real_asyncio.sleep(delay, result=result)

    loop = _wrap_loop(loop)
    future = loop.create_future()
    handle = loop.call_later(delay, future.set_result, result)

    def cancel_callback(future):
        if future.cancelled():
            handle.cancel()

    future.add_done_callback(cancel_callback)
    return future


def _wrap_loop(loop=None):
    """
    In order to deal with asyncio event loop compatibility issues,
    use this function to wrap the loop parameter for functions
    that support it. For example, since python3.4 does not have the
    AbstractEventLoop.create_future() method, this helper function
    can be used to add a wrapper that implements the create_future
    method for python3.4.

    @type loop: asyncio.AbstractEventLoop (or compatible)
    @param loop: event loop
    @rtype: asyncio.AbstractEventLoop (or compatible)
    @return: event loop
    """
    # The default loop returned by _wrap_loop should be consistent
    # with global_event_loop, in order to avoid accidental registration
    # of callbacks with a loop that is not intended to run.
    if hasattr(loop, "_asyncio_wrapper"):
        return loop

    # This returns a running loop if it exists, and otherwise returns
    # a loop associated with the current thread.
    safe_loop = _safe_loop(create=loop is None)
    if safe_loop is not None and (loop is None or safe_loop._loop is loop):
        return safe_loop

    if safe_loop is None:
        msg = f"_wrap_loop argument '{loop}' not associated with thread '{threading.get_ident()}'"
    else:
        msg = f"_wrap_loop argument '{loop}' different frome loop '{safe_loop._loop}' already associated with thread '{threading.get_ident()}'"

    if portage._internal_caller:
        raise AssertionError(msg)

    # It's not known whether external API consumers will trigger this case,
    # so if it happens then emit a UserWarning before returning a temporary
    # AsyncioEventLoop instance.
    warnings.warn(msg, UserWarning, stacklevel=2)

    # We could possibly add a weak reference in _thread_weakrefs.loops when
    # safe_loop is None, but if safe_loop is not None, then there is a
    # collision in _thread_weakrefs.loops that would need to be resolved.
    return _AsyncioEventLoop(loop=loop)


def _safe_loop(create: Optional[bool] = True) -> Optional[_AsyncioEventLoop]:
    """
    Return an event loop that's safe to use within the current context.
    For portage internal callers or external API consumers calling from
    the main thread, this returns a globally shared event loop instance.

    For external API consumers calling from a non-main thread, an
    asyncio loop must be registered for the current thread, or else the
    asyncio.get_event_loop() function will raise an error like this:

      RuntimeError: There is no current event loop in thread 'Thread-1'.

    In order to avoid this RuntimeError, a loop will be automatically
    created like this:

      asyncio.set_event_loop(asyncio.new_event_loop())

    In order to avoid a ResourceWarning, automatically created loops
    are added to a WeakValueDictionary, and closed via an atexit hook
    if they still exist during exit for the current pid.

    @type create: bool
    @param create: Create a loop by default if a loop is not already associated
        with the current thread. If create is False, then return None if a loop
        is not already associated with the current thread.
    @rtype: AsyncioEventLoop or None
    @return: event loop instance, or None if the create parameter is False and
        a loop is not already associated with the current thread.
    """
    loop = _get_running_loop()
    if loop is not None:
        return loop

    thread_key = threading.get_ident()
    with _thread_weakrefs.lock:
        if _thread_weakrefs.pid != portage.getpid():
            _thread_weakrefs.pid = portage.getpid()
            _thread_weakrefs.mainloop = None
            _thread_weakrefs.loops = weakref.WeakValueDictionary()
        try:
            loop = _thread_weakrefs.loops[thread_key]
            if loop.is_closed():
                # Discard wrapped asyncio.run loop that was closed.
                del _thread_weakrefs.loops[thread_key]
                if loop is _thread_weakrefs.mainloop:
                    _thread_weakrefs.mainloop = None
                loop = None
                raise KeyError(thread_key)
        except KeyError:
            if not create:
                return None
            try:
                try:
                    _loop = _real_asyncio.get_running_loop()
                except AttributeError:
                    _loop = _real_asyncio.get_event_loop()
            except RuntimeError:
                _loop = _real_asyncio.new_event_loop()
                _real_asyncio.set_event_loop(_loop)
            loop = _thread_weakrefs.loops[thread_key] = _AsyncioEventLoop(loop=_loop)

    if (
        _thread_weakrefs.mainloop is None
        and threading.current_thread() is threading.main_thread()
    ):
        _thread_weakrefs.mainloop = loop

    return loop


def _get_running_loop():
    """
    This calls the real asyncio get_running_loop() and wraps that with
    portage's internal AsyncioEventLoop wrapper. If there is no running
    asyncio event loop but portage has a reference to another running
    loop in this thread, then use that instead.

    This behavior enables portage internals to use the real asyncio.run
    while remaining compatible with internal code that does not use the
    real asyncio.run.
    """
    try:
        _loop = _real_asyncio.get_running_loop()
    except RuntimeError:
        _loop = None

    with _thread_weakrefs.lock:
        if _thread_weakrefs.pid == portage.getpid():
            try:
                loop = _thread_weakrefs.loops[threading.get_ident()]
            except KeyError:
                pass
            else:
                if _loop is loop._loop:
                    return loop
                elif _loop is None:
                    return loop if loop.is_running() else None

        if _loop is None:
            return None

        # If _loop it not None here it means it was probably a temporary
        # loop created by asyncio.run. Still keep a weak reference in case
        # we need to lookup this _AsyncioEventLoop instance later to add
        # _coroutine_exithandlers in the atexit_register function.
        if _thread_weakrefs.pid != portage.getpid():
            _thread_weakrefs.pid = portage.getpid()
            _thread_weakrefs.mainloop = None
            _thread_weakrefs.loops = weakref.WeakValueDictionary()

        loop = _thread_weakrefs.loops[threading.get_ident()] = _AsyncioEventLoop(
            loop=_loop
        )

        return loop


def _thread_weakrefs_atexit():
    while True:
        loop = None
        thread_key = None
        restore_loop = None
        with _thread_weakrefs.lock:
            if _thread_weakrefs.pid != portage.getpid():
                return

            try:
                thread_key, loop = _thread_weakrefs.loops.popitem()
            except KeyError:
                return
            else:
                # Temporarily associate it as the loop for the current thread so
                # that it can be looked up during run_coroutine_exitfuncs calls.
                # Also create a reference to a different loop if one is associated
                # with this thread so we can restore it later.
                try:
                    restore_loop = _thread_weakrefs.loops[threading.get_ident()]
                except KeyError:
                    pass
                _thread_weakrefs.loops[threading.get_ident()] = loop

        # Release the lock while closing the loop, since it may call
        # run_coroutine_exitfuncs interally.
        if loop is not None:
            loop.close()
            with _thread_weakrefs.lock:
                try:
                    if _thread_weakrefs.loops[threading.get_ident()] is loop:
                        del _thread_weakrefs.loops[threading.get_ident()]
                except KeyError:
                    pass
                if restore_loop is not None:
                    _thread_weakrefs.loops[threading.get_ident()] = restore_loop


_thread_weakrefs = types.SimpleNamespace(
    lock=threading.Lock(), loops=None, mainloop=None, pid=None
)
atexit_register(_thread_weakrefs_atexit)



import os
import subprocess
import sys
import unicodedata
from subprocess import PIPE as _PIPE, Popen as _Popen

import jaraco.envs


class VirtualEnv(jaraco.envs.VirtualEnv):
    name = '.env'
    # Some version of PyPy will import distutils on startup, implicitly
    # importing setuptools, and thus leading to BackendInvalid errors
    # when upgrading Setuptools. Bypass this behavior by avoiding the
    # early availability and need to upgrade.
    create_opts = ['--no-setuptools']

    def run(self, cmd, *args, **kwargs):
        cmd = [self.exe(cmd[0])] + cmd[1:]
        kwargs = {"cwd": self.root, "encoding": "utf-8", **kwargs}  # Allow overriding
        # In some environments (eg. downstream distro packaging), where:
        # - tox isn't used to run tests and
        # - PYTHONPATH is set to point to a specific setuptools codebase and
        # - no custom env is explicitly set by a test
        # PYTHONPATH will leak into the spawned processes.
        # In that case tests look for module in the wrong place (on PYTHONPATH).
        # Unless the test sets its own special env, pass a copy of the existing
        # environment with removed PYTHONPATH to the subprocesses.
        if "env" not in kwargs:
            env = dict(os.environ)
            if "PYTHONPATH" in env:
                del env["PYTHONPATH"]
            kwargs["env"] = env
        return subprocess.check_output(cmd, *args, **kwargs)


def _which_dirs(cmd):
    result = set()
    for path in os.environ.get('PATH', '').split(os.pathsep):
        filename = os.path.join(path, cmd)
        if os.access(filename, os.X_OK):
            result.add(path)
    return result


def run_setup_py(cmd, pypath=None, path=None, data_stream=0, env=None):
    """
    Execution command for tests, separate from those used by the
    code directly to prevent accidental behavior issues
    """
    if env is None:
        env = dict()
        for envname in os.environ:
            env[envname] = os.environ[envname]

    # override the python path if needed
    if pypath is not None:
        env["PYTHONPATH"] = pypath

    # override the execution path if needed
    if path is not None:
        env["PATH"] = path
    if not env.get("PATH", ""):
        env["PATH"] = _which_dirs("tar").union(_which_dirs("gzip"))
        env["PATH"] = os.pathsep.join(env["PATH"])

    cmd = [sys.executable, "setup.py"] + list(cmd)

    # https://bugs.python.org/issue8557
    shell = sys.platform == 'win32'

    try:
        proc = _Popen(
            cmd,
            stdout=_PIPE,
            stderr=_PIPE,
            shell=shell,
            env=env,
            encoding="utf-8",
        )

        if isinstance(data_stream, tuple):
            data_stream = slice(*data_stream)
        data = proc.communicate()[data_stream]
    except OSError:
        return 1, ''

    # decode the console string if needed
    if hasattr(data, "decode"):
        # use the default encoding
        data = data.decode()
        data = unicodedata.normalize('NFC', data)

    # communicate calls wait()
    return proc.returncode, data



# Copyright 2018-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import signal

import asyncio as _real_asyncio
from asyncio.events import AbstractEventLoop as _AbstractEventLoop

try:
    from asyncio.unix_events import _ThreadedChildWatcher as ThreadedChildWatcher
except ImportError:
    from asyncio.unix_events import ThreadedChildWatcher

try:
    from asyncio.unix_events import _PidfdChildWatcher as PidfdChildWatcher
except ImportError:
    from asyncio.unix_events import PidfdChildWatcher

import portage


class AsyncioEventLoop(_AbstractEventLoop):
    """
    Implementation of asyncio.AbstractEventLoop which wraps asyncio's
    event loop and is minimally compatible with _PortageEventLoop.
    """

    def __init__(self, loop=None):
        loop = loop or _real_asyncio.get_event_loop()
        self._loop = loop
        self.run_until_complete = self._run_until_complete
        self.call_soon = loop.call_soon
        self.call_soon_threadsafe = loop.call_soon_threadsafe
        self.call_later = loop.call_later
        self.call_at = loop.call_at
        self.is_running = loop.is_running
        self.is_closed = loop.is_closed
        self.close = self._close
        self.create_future = (
            loop.create_future
            if hasattr(loop, "create_future")
            else self._create_future
        )
        self.create_task = loop.create_task
        self.add_reader = loop.add_reader
        self.remove_reader = loop.remove_reader
        self.add_writer = loop.add_writer
        self.remove_writer = loop.remove_writer
        self.run_in_executor = loop.run_in_executor
        self.time = loop.time
        self.default_exception_handler = loop.default_exception_handler
        self.call_exception_handler = loop.call_exception_handler
        self.set_debug = loop.set_debug
        self.get_debug = loop.get_debug
        self._wakeup_fd = -1
        self._child_watcher = None
        # Used to drop recursive calls to _close.
        self._closing = False
        self._coroutine_exithandlers = []

        if portage._internal_caller:
            loop.set_exception_handler(self._internal_caller_exception_handler)

    def _close(self):
        """
        Before closing the main loop, run portage.process.run_exitfuncs()
        with the event loop running so that anything attached can clean
        itself up (like the socks5 ProxyManager for bug 925240).
        """
        if not (self._closing or self.is_closed()):
            self._closing = True
            if self._coroutine_exithandlers:
                self.run_until_complete(portage.process.run_coroutine_exitfuncs())
            self._loop.close()
            self._closing = False

    @staticmethod
    def _internal_caller_exception_handler(loop, context):
        """
        An exception handler which drops to a pdb shell if std* streams
        refer to a tty, and otherwise kills the process with SIGTERM.

        In order to avoid potential interference with API consumers, this
        implementation is only used when portage._internal_caller is True.
        """
        loop.default_exception_handler(context)
        if "exception" in context:
            # Normally emerge will wait for all coroutines to complete
            # after SIGTERM has been received. However, an unhandled
            # exception will prevent the interrupted coroutine from
            # completing, therefore use the default SIGTERM handler
            # in order to ensure that emerge exits immediately (though
            # uncleanly).
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
            os.kill(portage.getpid(), signal.SIGTERM)

    def _create_future(self):
        """
        Provide AbstractEventLoop.create_future() for python3.4.
        """
        return _real_asyncio.Future(loop=self._loop)

    @property
    def _asyncio_child_watcher(self):
        """
        Portage internals use this as a layer of indirection for
        asyncio.get_child_watcher(), in order to support versions of
        python where asyncio is not available.

        @rtype: asyncio.AbstractChildWatcher
        @return: the internal event loop's AbstractChildWatcher interface
        """
        if self._child_watcher is None:
            pidfd_works = False
            if PidfdChildWatcher is not None and hasattr(os, "pidfd_open"):
                try:
                    fd = os.pidfd_open(portage.getpid())
                except Exception:
                    pass
                else:
                    os.close(fd)
                    pidfd_works = True

            if pidfd_works:
                watcher = PidfdChildWatcher()
            else:
                watcher = ThreadedChildWatcher()

            if hasattr(watcher, "attach_loop"):
                watcher.attach_loop(self._loop)
            self._child_watcher = _ChildWatcherThreadSafetyWrapper(self, watcher)

        return self._child_watcher

    @property
    def _asyncio_wrapper(self):
        """
        Portage internals use this as a layer of indirection in cases
        where a wrapper around an asyncio.AbstractEventLoop implementation
        is needed for purposes of compatibility.

        @rtype: asyncio.AbstractEventLoop
        @return: the internal event loop's AbstractEventLoop interface
        """
        return self

    def _run_until_complete(self, future):
        """
        An implementation of AbstractEventLoop.run_until_complete that suppresses
        spurious error messages like the following reported in bug 655656:

            Exception ignored when trying to write to the signal wakeup fd:
            BlockingIOError: [Errno 11] Resource temporarily unavailable

        In order to avoid potential interference with API consumers, this
        implementation is only used when portage._internal_caller is True.
        """
        if not portage._internal_caller:
            return self._loop.run_until_complete(future)

        if self._wakeup_fd != -1:
            signal.set_wakeup_fd(self._wakeup_fd)
            self._wakeup_fd = -1
            # Account for any signals that may have arrived between
            # set_wakeup_fd calls.
            os.kill(portage.getpid(), signal.SIGCHLD)
        try:
            return self._loop.run_until_complete(future)
        finally:
            try:
                self._wakeup_fd = signal.set_wakeup_fd(-1)
            except ValueError:
                # This is intended to fail when not called in the main thread.
                pass


class _ChildWatcherThreadSafetyWrapper:
    """
    This class provides safety if multiple loops are running in different threads.
    """

    def __init__(self, loop, real_watcher):
        self._loop = loop
        self._real_watcher = real_watcher

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, a, b, c):
        pass

    def _child_exit(self, pid, status, callback, *args):
        self._loop.call_soon_threadsafe(callback, pid, status, *args)

    def add_child_handler(self, pid, callback, *args):
        self._real_watcher.add_child_handler(pid, self._child_exit, callback, *args)

    def remove_child_handler(self, pid):
        return self._real_watcher.remove_child_handler(pid)



from __future__ import annotations

import logging
import warnings

from typing import Any
from typing import Callable

import setuptools

from .. import _types as _t
from .pyproject_reading import PyProjectData
from .pyproject_reading import read_pyproject
from .setup_cfg import SetuptoolsBasicData
from .setup_cfg import extract_from_legacy
from .toml import InvalidTomlError
from .version_inference import get_version_inference_config

log = logging.getLogger(__name__)


def _warn_on_old_setuptools(_version: str = setuptools.__version__) -> None:
    if int(_version.split(".")[0]) < 61:
        warnings.warn(
            RuntimeWarning(
                f"""
ERROR: setuptools=={_version} is used in combination with setuptools-scm>=8.x

Your build configuration is incomplete and previously worked by accident!
setuptools-scm requires setuptools>=61 (recommended: >=80)

Suggested workaround if applicable:
 - migrating from the deprecated setup_requires mechanism to pep517/518
   and using a pyproject.toml to declare build dependencies
   which are reliably pre-installed before running the build tools
"""
            )
        )


_warn_on_old_setuptools()


def _log_hookstart(hook: str, dist: setuptools.Distribution) -> None:
    log.debug(
        "%s %s %s %r",
        hook,
        id(dist),
        id(dist.metadata),
        {**vars(dist.metadata), "long_description": ...},
    )


def get_keyword_overrides(
    value: bool | dict[str, Any] | Callable[[], dict[str, Any]],
) -> dict[str, Any]:
    """normalize the version keyword input"""
    if value is True:
        return {}
    elif callable(value):
        return value()
    else:
        assert isinstance(value, dict), "version_keyword expects a dict or True"
        return value


def version_keyword(
    dist: setuptools.Distribution,
    keyword: str,
    value: bool | dict[str, Any] | Callable[[], dict[str, Any]],
    *,
    _given_pyproject_data: _t.GivenPyProjectResult = None,
    _given_legacy_data: SetuptoolsBasicData | None = None,
    _get_version_inference_config: _t.GetVersionInferenceConfig = get_version_inference_config,
) -> None:
    """apply version infernce when setup(use_scm_version=...) is used
    this takes priority over the finalize_options based version
    """

    _log_hookstart("version_keyword", dist)

    # Parse overrides (integration point responsibility)
    overrides = get_keyword_overrides(value)

    assert "dist_name" not in overrides, (
        "dist_name may not be specified in the setup keyword "
    )

    legacy_data = extract_from_legacy(dist, _given_legacy_data=_given_legacy_data)
    dist_name: str | None = legacy_data.name

    was_set_by_infer = getattr(dist, "_setuptools_scm_version_set_by_infer", False)

    # Exit early if overrides is empty dict AND version was set by infer
    if overrides == {} and was_set_by_infer:
        return

    # Get pyproject data (support direct injection for tests)
    try:
        pyproject_data = read_pyproject(_given_result=_given_pyproject_data)
    except FileNotFoundError:
        log.debug("pyproject.toml not found, proceeding with empty configuration")
        pyproject_data = PyProjectData.empty()
    except InvalidTomlError as e:
        log.debug("Configuration issue in pyproject.toml: %s", e)
        return

    # Pass None as current_version if overrides is truthy AND version was set by infer
    current_version = (
        None
        if (overrides and was_set_by_infer)
        else (legacy_data.version or pyproject_data.project_version)
    )

    result = _get_version_inference_config(
        dist_name=dist_name,
        current_version=current_version,
        pyproject_data=pyproject_data,
        overrides=overrides,
    )

    result.apply(dist)


def infer_version(
    dist: setuptools.Distribution,
    *,
    _given_pyproject_data: _t.GivenPyProjectResult = None,
    _given_legacy_data: SetuptoolsBasicData | None = None,
    _get_version_inference_config: _t.GetVersionInferenceConfig = get_version_inference_config,
) -> None:
    """apply version inference from the finalize_options hook
    this is the default for pyproject.toml based projects that don't use the use_scm_version keyword

    if the version keyword is used, it will override the version from this hook
    as user might have passed custom code version schemes
    """

    _log_hookstart("infer_version", dist)

    legacy_data = extract_from_legacy(dist, _given_legacy_data=_given_legacy_data)
    dist_name = legacy_data.name

    try:
        pyproject_data = read_pyproject(_given_result=_given_pyproject_data)
    except FileNotFoundError:
        log.debug("pyproject.toml not found, skipping infer_version")
        return
    except InvalidTomlError as e:
        log.debug("Configuration issue in pyproject.toml: %s", e)
        return

    # Only infer when tool section present per get_version_inference_config
    result = _get_version_inference_config(
        dist_name=dist_name,
        current_version=legacy_data.version or pyproject_data.project_version,
        pyproject_data=pyproject_data,
    )
    result.apply(dist)
